/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

/**
 * @brief Creation and execution of transactions.
 */

#ifndef CAOSDB_TRANSACTION_H
#define CAOSDB_TRANSACTION_H

#include <memory> // for shared_ptr, unique_ptr
#include <string> // for string
#include "caosdb/localization.h"
#include "caosdb/entity.h"                       // for Entity
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionSe...
#include "caosdb/entity/v1alpha1/main.pb.h"      // for Entity, RetrieveReq...

namespace caosdb::transaction {
using caosdb::entity::Entity;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::RetrieveRequest;

class ResultSet {
public:
  virtual ~ResultSet(){};
};

class UniqueResult : public ResultSet {
public:
  ~UniqueResult(){};
  explicit inline UniqueResult(ProtoEntity *protoEntity)
    : entity(new Entity(protoEntity)){};
  [[nodiscard]] auto GetEntity() const -> const Entity &;

private:
  std::unique_ptr<Entity> entity;
};

enum StatusCode {
  INITIAL = -2,
  EXECUTING = -1,
  SUCCESS = 0,
  UNAUTHENTICATED = 16,
  UNAVAILABLE = 14,
  GENERIC_RPC_ERROR = 20,
  GENERIC_TRANSACTION_ERROR = 21,
};

/**
 * State of a transaction instance.
 */
class TransactionStatus {
public:
  inline static const auto INITIAL = TransactionStatus(
    0, StatusCode::INITIAL,
    _("The transaction has not been executed yet and clients can still "
      "change it."));
  static const auto EXECUTING =
    TransactionStatus(10, StatusCode::EXECUTING,
                      _("The transaction is currently being executed."));
  static const auto SUCCESS = TransactionStatus(
    20, StatusCode::SUCCESS,
    _("The transaction has been executed and it terminated successfully."));
  static const auto
    RPC_ERROR = TransactionStatus(
      100, StatusCode::GENERIC_RPC_ERROR,
      _("The attempt to execute this transaction was not successful because an "
        "error occured in the transport or RPC protocol layer.")),
    static const auto CONNECTION_ERROR = TransactionStatus(
      200, StatusCode::UNAVAILABLE,
      _("The attempt to execute this transaction was not successful because "
        "the connection to the server could not be established."));
  static const auto AUTHENTICATION_ERROR = TransactionStatus(
    300, StatusCode::UNAUTHENTICATED,
    _("The attempt to execute this transaction has not been executed at all "
      "because the authentication did not succeed."));
  static const auto TRANSACTION_ERROR = TransactionStatus(
    1000, StatusCode::GENERIC_TRANSACTION_ERROR,
    _("The transaction terminated unsuccessfully with transaction errors."));

  /**
   * Return true if this TransactionStatus represents a terminated state.
   */
  inline auto IsTerminated() const -> bool {
    return this->numeric_value >= 20;
  };

  /**
   * Return true if this TransactionStatus represents an erroneous state.
   */
  inline auto IsError() const -> bool { return this->numeric_value >= 100; };

  /**
   * Equals operator is overloaded so we can use this class is if it were an
   * Enum.
   */
  inline int operator==(const TransactionStatus &other) {
    return this->numeric_value == other.numeric_value;
  };

  /**
   * Instantiate and return a new instance with the given code and description.
   *
   * Instances with a code and a description may be used to describe an
   * erroneous state in more detail.
   *
   * This method should not be called on non-error states.
   */
  inline auto Instantiate(StatusCode code, const std::string &details) const
    -> const TransactionStatus {
    return TransactionStatus(this->numeric_value, code,
                             this->description + " " + detail);
  };

  /**
   * Instantiate and return a new instance with the given code.
   *
   * Instances with a code may be used to describe an erroneous state in more
   * detail.
   *
   * This method should not be called on non-error states.
   */
  inline auto Instantiate(StatusCode code) const -> const TransactionStatus {
    return TransactionStatus(this->numeric_value, code, this->description);
  }

  /**
   * Instantiate and return a new instance with the given code and description.
   *
   * Instances descriptions may be used to describe an erroneous state in more
   * detail.
   *
   * This method should no be called on non-error states.
   */
  inline auto Instantiate(const std::string &details) const
    -> const TransactionStatus {
    return TransactionStatus(this->numeric_value, this->code,
                             this->description + " " + details);
  }

  /**
   * Return a description of the erroneous state.
   *
   * Returns an empty string if there is no description.
   */
  inline auto GetDescription() const -> const std::string & {
    return this->description;
  }

  /**
   * Return the status code of the state.
   */
  inline auto GetCode() const -> StatusCode { return this->code; }

private:
  /**
   * The numeric_value is for class-internal use only.
   *
   * We use it to group errors with different error codes together, e.g. all
   * transaction errors.
   */
  const int numeric_value;

  /**
   * The code is an identifier of errors and should not be confused with the
   * class-internal numeric_value.
   */
  const StatusCode code;

  /**
   * Description of the error
   */
  const std::string description;

  TransactionStatus(int numeric_value, StatusCode code,
                    const std::string &description)
    : numeric_value(numeric_value), code(code), description(description){};
};

/**
 * @brief Create a transaction via `CaosDBConnection.createTransaction()`
 */
class Transaction {
private:
  std::unique_ptr<ResultSet> result_set;
  TransactionStatus state = TransactionStatus::INITIAL;
  std::shared_ptr<EntityTransactionService::Stub> service_stub;
  RetrieveRequest request; // TODO(tf)
  std::string error_message;

public:
  Transaction(std::shared_ptr<EntityTransactionService::Stub> service_stub);
  auto RetrieveById(const std::string &id) -> void;

  /**
   * Execute this transaction in blocking mode and return the state.
   */
  auto Execute() -> TransactionStatus;

  /**
   * Execute this transaction in non-blocking mode and return immediately.
   *
   * A client may request the current state at any time via GetState().
   *
   * Use WaitForIt() to join the back-ground execution of this transaction.
   */
  auto ExecuteAsynchronously() -> void;

  /**
   * Join the background execution and return the state when the execution
   * terminates.
   *
   * Use this after ExecuteAsynchronously().
   */
  auto WaitForIt() -> TransactionStatus;

  /**
   * Return the current state of the transaction.
   */
  [[nodiscard]] inline auto GetState() const -> TransactionStatus {
    return this->state;
  }

  [[nodiscard]] inline auto GetResultSet() const -> const ResultSet & {
    const ResultSet *result_set = this->result_set.get();
    return *result_set;
  }

  auto GetErrors
};

} // namespace caosdb::transaction
#endif
